'''Provides tools for building models.'''
from logging import info
from os import system, walk, makedirs
from os.path import exists, basename, dirname
from multiprocessing import Pool
from glob import glob
from hashlib import md5
from shutil import copyfile, move, rmtree
from ya2.build.mtprocesser import ProcesserMgr
from ya2.build.build import to_be_built


class ModelsBuilder():

    def __init__(self):
        self._cache_files = []  # for avoiding rebuilding the same file

    def build(self, blend_path, cores):
        '''Builds the models i.e. creates glTF and bam files from blend
        ones.'''
        args = []
        for root, _, fnames in walk(blend_path):
            for fname in [fname for fname in fnames if fname.endswith('.blend')]:
                if '/prototypes/' not in root:
                    args += [(root, fname)]
        with Pool() as p:
            p.starmap(self._export_blend, args)
        # caching is broken now: it is based on the previous
        # non-multiprocessing approach
        # i should evaluate if fix it or just change the blend_path for building
        # only specific models
        cache, lines = [], []
        if exists('hash_cache.txt'):
            with open('hash_cache.txt') as fhash:
                lines = fhash.readlines()
        for line in lines:  # line's e.g. assets/path/to/gltf_or_png.ext 68ced1
            line_spl = line.split()
            hashval = line_spl[-1]
            fname = ' '.join(line_spl[:-1])
            if fname not in self._cache_files:
                cache += [(fname, hashval)]
        for cfile in self._cache_files:
            cache += [(cfile, md5(open(cfile, 'rb').read()).hexdigest())]
        cache = cache[-2048:]
        with open('hash_cache.txt', 'w') as fhash:
            fhash.write('\n'.join([' '.join(line) for line in cache]))

    def _export_blend(self, root, fname):
        '''Exports blend files to glTF and bam formats.'''
        if self._export_gltf(root, fname):
            self._export_bam(root, fname)

    def _export_gltf(self, root, fname):
        '''Exports glTF files from blend ones.'''
        _fname = '%s/%s' % (root, fname)
        pgltf = 'assets/models/gltf/'
        new_dir = root.replace('assets/models/blend/', pgltf)
        rmtree(new_dir, ignore_errors=True)
        makedirs(new_dir)
        #files_before = [basename(gname) for gname in glob('./*')]
        cmd = 'blender %s --background --python ya2/build/blend2gltf.py '
        cmd += '-- %s.gltf'
        cmd = cmd % (_fname, new_dir + '/' + fname[:-6])
        gltf_name = _fname.replace('assets/models/blend/', pgltf)
        gltf_name = gltf_name.replace('.blend', '.gltf')
        if not to_be_built(gltf_name, [_fname]):
            return False
        system(cmd)
        self._cache_files += [_fname, gltf_name]
        #files_after = [basename(gname) for gname in glob('./*')]
        #new_files = [nnm for nnm in files_after if nnm not in files_before]
        #new_dir = root.replace('assets/models/blend/', pgltf)
        #rmtree(new_dir, ignore_errors=True)
        #makedirs(new_dir)
        #for mname in new_files:
        #    new_name = '%s/%s' % (new_dir, mname)
        #    move(mname, new_name)
        #    info('move %s %s' % (mname, new_name))
        # # blender rewrites metal files: let's restore them
        # metal_files = [fnm for fnm in glob(new_dir + '/*') if 'metal' in fnm]
        # for metal_file in metal_files:
        #     src = metal_file.replace(pgltf, 'assets/models/')
        #     if not exists(src):
        #         src = metal_file.replace(pgltf, 'assets/models/prototypes/')
        #         src_split = src.split('/')
        #         src_tracks_idx = src_split.index('tracks')
        #         before = src_split[:src_tracks_idx]
        #         after = src_split[src_tracks_idx + 2:]
        #         src = '/'.join(before + after)
        #     copyfile(src, metal_file)
        return True

    def _export_bam(self, root, fname):
        '''Exports bam files from glTF ones.'''
        _fname = '%s/%s' % (root, fname)
        gltf_name = (_fname[:-5] + 'gltf').replace('/blend/', '/gltf/', 1)
        bam_name = (_fname[:-5] + 'bam').replace('/blend/', '/bam/', 1)
        cmd_args = gltf_name, bam_name
        # use dds files in place of png/jpg in gltf2bam
        copyfile(gltf_name, gltf_name + '.tmp')
        with open(gltf_name) as fgltf:
            lines = fgltf.readlines()
        deps = []
        for line in lines:
            if ('.png' in line or '.jpg' in line) and '"uri"' in line:
                rln = line[line.index('"uri"') + 9:].rstrip(',\n"')
                tname = '%s/%s' % (root, rln)
                deps += [tname.replace('/models/blend/', '/models/gltf/', 1)]
        for dep in deps:
            tgt = dep.replace('/gltf/', '/bam/', 1)
            tgt = tgt.replace('.png', '.dds').replace('.jpg', '.dds')
            makedirs(dirname(tgt), exist_ok=True)
            info('convert %s %s' % (dep, tgt))
            system('convert %s %s' % (dep, tgt))
        rpl = lambda lin: lin.replace('.png', '.dds').replace('.jpg', '.dds').replace('/png', '/dds').replace('/jpg', '/dds')
        with open(gltf_name, 'w') as fgltf:
            fgltf.write(''.join([rpl(line) for line in lines]))
        makedirs(dirname(bam_name), exist_ok=True)
        if to_be_built(bam_name, deps):
            system('gltf2bam %s %s' % cmd_args)
        self._cache_files += [bam_name] + deps
