'''A multi-threaded command dispatcher.'''
from datetime import datetime
from multiprocessing import cpu_count
from threading import Thread, RLock
from os import system
from logging import info, debug


class ProcesserThread(Thread):
    '''A thread which asynchronously processes commands from a command list.'''

    def __init__(self, cmd_lst, lock):
        '''The constructor.
        cmd_lst: commands that can be exectuted from serveral processers
        lock: shared lock between processers for accessing the command list'''
        Thread.__init__(self)
        self.cmd_lst = cmd_lst
        self.lock = lock

    def run(self):
        '''The thread's logics.'''
        while True:
            with self.lock:
                if not self.cmd_lst:
                    return
                cmd = self.cmd_lst.pop(0)
            info('%s %s' % (datetime.now().strftime("%H:%M:%S"), cmd))
            system(cmd)


class SyncProcesser:
    '''Synchronously processes a command list.'''

    def __init__(self, cmd_lst):
        '''The constructor.
        cmd_lst: the list that must be executed'''
        self.cmd_lst = cmd_lst

    def run(self):
        '''The processer's logics.'''
        for cmd in self.cmd_lst:
            before_str = datetime.now().strftime("(executing) %H:%M:%S")
            info('%s %s' % (before_str, cmd))
            system(cmd)
            after_str = datetime.now().strftime("(executed) %H:%M:%S")
            debug('%s %s' % (after_str, cmd))


class ProcesserMgr:
    '''Synchronously processes commands that are submitted (eventually) using
    multiple threads.'''

    def __init__(self, cores):
        '''The constructor.
        cores: how many cpu cores are used to process the commands'''
        try:
            self.cores = cpu_count()
        except NotImplementedError:
            self.cores = 1
        self.cores = cores if cores else int(self.cores / 4 + 1)
        debug('processer-mgr: using %s cores' % self.cores)
        self.cmd_lst = []

    def add(self, cmd):
        '''Adds cmd to the list that will be processed.'''
        self.cmd_lst += [cmd]

    def run(self):
        '''Performs the commands that have been added.'''
        if self.cores != 1:
            threads, lock = [], RLock()
            threads = [ProcesserThread(self.cmd_lst, lock)
                       for _ in range(self.cores)]
            [thread.start() for thread in threads]
            [thread.join() for thread in threads]
        else:
            SyncProcesser(self.cmd_lst).run()
